SANIC_APP = "srv"
BASE_PATH_ENV = "SRV_BASE_PATH"
LOGGER = "srv"
SETTINGS_MODULE = "server_conf.settings"
SETTINGS_MODULE_VAR = "SRV_SETTINGS_MODULE"
SUPER_USER_GROUP = "root"
GOOGLE_AUTH_ENV = "GOOGLE_APPLICATION_CREDENTIALS"
