{% extends "_layout.html" %}
{% block content %}
  <div id="app"></div>
  <p class="text-indigo-500 text-4xl">
    {% if data.content.username %}
    Hello {{ data.content.username }}
    {% else %}
    Hello Anon user..
    {% endif %}
    {{ data }}
  </p>
{% endblock content %}

